@@ -5,18 +5,19 @@
 // exchangers to check that incoming mail from a domain comes from a host
 // authorized by that domain's administrators [Wikipedia].
 //
-// This is a Go implementation of it, which is used by the chasquid SMTP
-// server (https://blitiri.com.ar/p/chasquid/).
+// This package is intended to be used by SMTP servers to implement SPF
+// validation.
 //
-// Supported mechanisms and modifiers:
+// All mechanisms and modifiers are supported:
 //   all
 //   include
 //   a
 //   mx
+//   ptr
 //   ip4
 //   ip6
-//   redirect
 //   exists
+//   redirect
 //   exp (ignored)
 //   Macros
 //
@@ -26,6 +27,8 @@
 package spf // import "blitiri.com.ar/go/spf"
 
 import (
+	"context"
+	"errors"
 	"fmt"
 	"net"
 	"net/url"
@@ -34,15 +37,6 @@ import (
 	"strings"
 )
 
-// Functions that we can override for testing purposes.
-var (
-	lookupTXT  = net.LookupTXT
-	lookupMX   = net.LookupMX
-	lookupIP   = net.LookupIP
-	lookupAddr = net.LookupAddr
-	trace      = func(f string, a ...interface{}) {}
-)
-
 // The Result of an SPF check. Note the values have meaning, we use them in
 // headers.  https://tools.ietf.org/html/rfc7208#section-8
 type Result string
@@ -85,52 +79,171 @@ var qualToResult = map[byte]Result{
 	'?': Neutral,
 }
 
+// Errors returned by the library. Note that the errors returned in different
+// situations may change over time, and new ones may be added. Be careful
+// about over-relying on these.
 var (
-	errLookupLimitReached = fmt.Errorf("lookup limit reached")
-	errUnknownField       = fmt.Errorf("unknown field")
-	errInvalidIP          = fmt.Errorf("invalid ipX value")
-	errInvalidMask        = fmt.Errorf("invalid mask")
-	errInvalidMacro       = fmt.Errorf("invalid macro")
-	errInvalidDomain      = fmt.Errorf("invalid domain")
-	errNoResult           = fmt.Errorf("no DNS record found")
-	errMultipleRecords    = fmt.Errorf("multiple matching DNS records")
-	errTooManyMXRecords   = fmt.Errorf("too many MX records")
+	// Errors related to an invalid SPF record.
+	ErrUnknownField  = errors.New("unknown field")
+	ErrInvalidIP     = errors.New("invalid ipX value")
+	ErrInvalidMask   = errors.New("invalid mask")
+	ErrInvalidMacro  = errors.New("invalid macro")
+	ErrInvalidDomain = errors.New("invalid domain")
 
-	errMatchedAll    = fmt.Errorf("matched 'all'")
-	errMatchedA      = fmt.Errorf("matched 'a'")
-	errMatchedIP     = fmt.Errorf("matched 'ip'")
-	errMatchedMX     = fmt.Errorf("matched 'mx'")
-	errMatchedPTR    = fmt.Errorf("matched 'ptr'")
-	errMatchedExists = fmt.Errorf("matched 'exists'")
+	// Errors related to DNS lookups.
+	// Note that the library functions may also return net.DNSError.
+	ErrNoResult           = errors.New("no DNS record found")
+	ErrLookupLimitReached = errors.New("lookup limit reached")
+	ErrTooManyMXRecords   = errors.New("too many MX records")
+	ErrMultipleRecords    = errors.New("multiple matching DNS records")
+
+	// Errors returned on a successful match.
+	ErrMatchedAll    = errors.New("matched all")
+	ErrMatchedA      = errors.New("matched a")
+	ErrMatchedIP     = errors.New("matched ip")
+	ErrMatchedMX     = errors.New("matched mx")
+	ErrMatchedPTR    = errors.New("matched ptr")
+	ErrMatchedExists = errors.New("matched exists")
 )
 
+// Default value for the maximum number of DNS lookups while resolving SPF.
+// RFC is quite clear 10 must be the maximum allowed.
+// https://tools.ietf.org/html/rfc7208#section-4.6.4
+const defaultMaxLookups = 10
+
+// TraceFunc is the type of tracing functions.
+type TraceFunc func(f string, a ...interface{})
+
+var (
+	nullTrace    = func(f string, a ...interface{}) {}
+	defaultTrace = nullTrace
+)
+
+// Option type, for setting options. Users are expected to treat this as an
+// opaque type and not rely on the implementation, which is subject to change.
+type Option func(*resolution)
+
 // CheckHost fetches SPF records for `domain`, parses them, and evaluates them
 // to determine if `ip` is permitted to send mail for it.
 // Because it doesn't receive enough information to handle macros well, its
 // usage is not recommended, but remains supported for backwards
 // compatibility.
+//
+// The function returns a Result, which corresponds with the SPF result for
+// the check as per RFC, as well as an error for debugging purposes. Note that
+// the error may be non-nil even on successful checks.
+//
 // Reference: https://tools.ietf.org/html/rfc7208#section-4
+//
+// Deprecated: use CheckHostWithSender instead.
 func CheckHost(ip net.IP, domain string) (Result, error) {
-	trace("check host %q %q", ip, domain)
-	r := &resolution{ip, 0, "@" + domain, nil}
+	r := &resolution{
+		ip:       ip,
+		maxcount: defaultMaxLookups,
+		helo:     domain,
+		sender:   "@" + domain,
+		ctx:      context.TODO(),
+		resolver: defaultResolver,
+		trace:    defaultTrace,
+	}
 	return r.Check(domain)
 }
 
 // CheckHostWithSender fetches SPF records for `sender`'s domain, parses them,
 // and evaluates them to determine if `ip` is permitted to send mail for it.
 // The `helo` domain is used if the sender has no domain part.
+//
+// The `opts` optional parameter can be used to adjust some specific
+// behaviours, such as the maximum number of DNS lookups allowed.
+//
+// The function returns a Result, which corresponds with the SPF result for
+// the check as per RFC, as well as an error for debugging purposes. Note that
+// the error may be non-nil even on successful checks.
+//
 // Reference: https://tools.ietf.org/html/rfc7208#section-4
-func CheckHostWithSender(ip net.IP, helo, sender string) (Result, error) {
+func CheckHostWithSender(ip net.IP, helo, sender string, opts ...Option) (Result, error) {
 	_, domain := split(sender)
 	if domain == "" {
 		domain = helo
 	}
 
-	trace("check host with sender %q %q %q (%q)", ip, helo, sender, domain)
-	r := &resolution{ip, 0, sender, nil}
+	r := &resolution{
+		ip:       ip,
+		maxcount: defaultMaxLookups,
+		helo:     helo,
+		sender:   sender,
+		ctx:      context.TODO(),
+		resolver: defaultResolver,
+		trace:    defaultTrace,
+	}
+
+	for _, opt := range opts {
+		opt(r)
+	}
+
 	return r.Check(domain)
 }
 
+// OverrideLookupLimit overrides the maximum number of DNS lookups allowed
+// during SPF evaluation. Note that using this violates the RFC, which is
+// quite explicit that the maximum allowed MUST be 10 (the default). Please
+// use with care.
+//
+// This is EXPERIMENTAL for now, and the API is subject to change.
+func OverrideLookupLimit(limit uint) Option {
+	return func(r *resolution) {
+		r.maxcount = limit
+	}
+}
+
+// WithContext is an option to set the context for this operation, which will
+// be passed along to the resolver functions and other external calls if
+// needed.
+//
+// This is EXPERIMENTAL for now, and the API is subject to change.
+func WithContext(ctx context.Context) Option {
+	return func(r *resolution) {
+		r.ctx = ctx
+	}
+}
+
+// DNSResolver implements the methods we use to resolve DNS queries.
+// It is intentionally compatible with *net.Resolver.
+type DNSResolver interface {
+	LookupTXT(ctx context.Context, name string) ([]string, error)
+	LookupMX(ctx context.Context, name string) ([]*net.MX, error)
+	LookupIPAddr(ctx context.Context, host string) ([]net.IPAddr, error)
+	LookupAddr(ctx context.Context, addr string) (names []string, err error)
+}
+
+var defaultResolver DNSResolver = net.DefaultResolver
+
+// WithResolver sets the resolver to use for DNS lookups. It can be useful for
+// testing, and for customize DNS resolution specifically for this library.
+//
+// The default is to use net.DefaultResolver, which should be appropriate for
+// most users.
+//
+// This is EXPERIMENTAL for now, and the API is subject to change.
+func WithResolver(resolver DNSResolver) Option {
+	return func(r *resolution) {
+		r.resolver = resolver
+	}
+}
+
+// WithTraceFunc sets the resolver's trace function.
+//
+// This can be used for debugging. The trace messages are NOT machine
+// parseable, and are NOT stable. They should also NOT be included in
+// user-visible output, as they may include sensitive details.
+//
+// This is EXPERIMENTAL for now, and the API is subject to change.
+func WithTraceFunc(trace TraceFunc) Option {
+	return func(r *resolution) {
+		r.trace = trace
+	}
+}
+
 // split an user@domain address into user and domain.
 func split(addr string) (string, string) {
 	ps := strings.SplitN(addr, "@", 2)
@@ -142,43 +255,54 @@ func split(addr string) (string, string) {
 }
 
 type resolution struct {
-	ip    net.IP
-	count uint
+	ip       net.IP
+	count    uint
+	maxcount uint
 
+	helo   string
 	sender string
 
 	// Result of doing a reverse lookup for ip (so we only do it once).
 	ipNames []string
+
+	// Context for this resolution.
+	ctx context.Context
+
+	// DNS resolver to use.
+	resolver DNSResolver
+
+	// Trace function, used for debugging.
+	trace TraceFunc
 }
 
-var aField = regexp.MustCompile(`^a$|a:|a/`)
-var mxField = regexp.MustCompile(`^mx$|mx:|mx/`)
-var ptrField = regexp.MustCompile(`^ptr$|ptr:`)
+var aField = regexp.MustCompile(`^(a$|a:|a/)`)
+var mxField = regexp.MustCompile(`^(mx$|mx:|mx/)`)
+var ptrField = regexp.MustCompile(`^(ptr$|ptr:)`)
 
 func (r *resolution) Check(domain string) (Result, error) {
 	r.count++
-	trace("check %s %d", domain, r.count)
-	txt, err := getDNSRecord(domain)
+	r.trace("check %q %d", domain, r.count)
+	txt, err := r.getDNSRecord(domain)
 	if err != nil {
 		if isTemporary(err) {
-			trace("dns temp error: %v", err)
+			r.trace("dns temp error: %v", err)
 			return TempError, err
 		}
-		if err == errMultipleRecords {
-			trace("multiple dns records")
+		if err == ErrMultipleRecords {
+			r.trace("multiple dns records")
 			return PermError, err
 		}
 		// Could not resolve the name, it may be missing the record.
 		// https://tools.ietf.org/html/rfc7208#section-2.6.1
-		trace("dns perm error: %v", err)
+		r.trace("dns perm error: %v", err)
 		return None, err
 	}
-	trace("dns record %q", txt)
+	r.trace("dns record %q", txt)
 
 	if txt == "" {
 		// No record => None.
 		// https://tools.ietf.org/html/rfc7208#section-4.5
-		return None, errNoResult
+		return None, ErrNoResult
 	}
 
 	fields := strings.Split(txt, " ")
@@ -196,7 +320,7 @@ func (r *resolution) Check(domain string) (Result, error) {
 	if len(redirects) > 1 {
 		// At most a single redirect is allowed.
 		// https://tools.ietf.org/html/rfc7208#section-6
-		return PermError, errInvalidDomain
+		return PermError, ErrInvalidDomain
 	}
 	fields = append(newfields, redirects...)
 
@@ -212,11 +336,11 @@ func (r *resolution) Check(domain string) (Result, error) {
 			continue
 		}
 
-		// Limit the number of resolutions to 10
+		// Limit the number of resolutions.
 		// https://tools.ietf.org/html/rfc7208#section-4.6.4
-		if r.count > 10 {
-			trace("lookup limit reached")
-			return PermError, errLookupLimitReached
+		if r.count > r.maxcount {
+			r.trace("lookup limit reached")
+			return PermError, ErrLookupLimitReached
 		}
 
 		// See if we have a qualifier, defaulting to + (pass).
@@ -234,54 +358,54 @@ func (r *resolution) Check(domain string) (Result, error) {
 
 		if lfield == "all" {
 			// https://tools.ietf.org/html/rfc7208#section-5.1
-			trace("%v matched all", result)
-			return result, errMatchedAll
+			r.trace("%v matched all", result)
+			return result, ErrMatchedAll
 		} else if strings.HasPrefix(lfield, "include:") {
 			if ok, res, err := r.includeField(result, field, domain); ok {
-				trace("include ok, %v %v", res, err)
+				r.trace("include ok, %v %v", res, err)
 				return res, err
 			}
 		} else if aField.MatchString(lfield) {
 			if ok, res, err := r.aField(result, field, domain); ok {
-				trace("a ok, %v %v", res, err)
+				r.trace("a ok, %v %v", res, err)
 				return res, err
 			}
 		} else if mxField.MatchString(lfield) {
 			if ok, res, err := r.mxField(result, field, domain); ok {
-				trace("mx ok, %v %v", res, err)
+				r.trace("mx ok, %v %v", res, err)
 				return res, err
 			}
 		} else if strings.HasPrefix(lfield, "ip4:") || strings.HasPrefix(lfield, "ip6:") {
 			if ok, res, err := r.ipField(result, field); ok {
-				trace("ip ok, %v %v", res, err)
+				r.trace("ip ok, %v %v", res, err)
 				return res, err
 			}
 		} else if ptrField.MatchString(lfield) {
 			if ok, res, err := r.ptrField(result, field, domain); ok {
-				trace("ptr ok, %v %v", res, err)
+				r.trace("ptr ok, %v %v", res, err)
 				return res, err
 			}
 		} else if strings.HasPrefix(lfield, "exists:") {
 			if ok, res, err := r.existsField(result, field, domain); ok {
-				trace("exists ok, %v %v", res, err)
+				r.trace("exists ok, %v %v", res, err)
 				return res, err
 			}
 		} else if strings.HasPrefix(lfield, "exp=") {
-			trace("exp= not used, skipping")
+			r.trace("exp= not used, skipping")
 			continue
 		} else if strings.HasPrefix(lfield, "redirect=") {
-			trace("redirect, %q", field)
+			r.trace("redirect, %q", field)
 			return r.redirectField(field, domain)
 		} else {
 			// http://www.openspf.org/SPF_Record_Syntax
-			trace("permerror, unknown field")
-			return PermError, errUnknownField
+			r.trace("permerror, unknown field")
+			return PermError, ErrUnknownField
 		}
 	}
 
 	// Got to the end of the evaluation without a result => Neutral.
 	// https://tools.ietf.org/html/rfc7208#section-4.7
-	trace("fallback to neutral")
+	r.trace("fallback to neutral")
 	return Neutral, nil
 }
 
@@ -290,8 +414,8 @@ func (r *resolution) Check(domain string) (Result, error) {
 // https://tools.ietf.org/html/rfc7208#section-3
 // https://tools.ietf.org/html/rfc7208#section-3.2
 // https://tools.ietf.org/html/rfc7208#section-4.5
-func getDNSRecord(domain string) (string, error) {
-	txts, err := lookupTXT(domain)
+func (r *resolution) getDNSRecord(domain string) (string, error) {
+	txts, err := r.resolver.LookupTXT(r.ctx, domain)
 	if err != nil {
 		return "", err
 	}
@@ -322,7 +446,7 @@ func getDNSRecord(domain string) (string, error) {
 	} else if l == 1 {
 		return records[0], nil
 	}
-	return "", errMultipleRecords
+	return "", ErrMultipleRecords
 }
 
 func isTemporary(err error) bool {
@@ -336,18 +460,18 @@ func (r *resolution) ipField(res Result, field string) (bool, Result, error) {
 	if strings.Contains(fip, "/") {
 		_, ipnet, err := net.ParseCIDR(fip)
 		if err != nil {
-			return true, PermError, errInvalidMask
+			return true, PermError, ErrInvalidMask
 		}
 		if ipnet.Contains(r.ip) {
-			return true, res, errMatchedIP
+			return true, res, ErrMatchedIP
 		}
 	} else {
 		ip := net.ParseIP(fip)
 		if ip == nil {
-			return true, PermError, errInvalidIP
+			return true, PermError, ErrInvalidIP
 		}
 		if ip.Equal(r.ip) {
-			return true, res, errMatchedIP
+			return true, res, ErrMatchedIP
 		}
 	}
 
@@ -364,17 +488,17 @@ func (r *resolution) ptrField(res Result, field, domain string) (bool, Result, e
 	}
 	ptrDomain, err := r.expandMacros(ptrDomain, domain)
 	if err != nil {
-		return true, PermError, errInvalidMacro
+		return true, PermError, ErrInvalidMacro
 	}
 
 	if ptrDomain == "" {
-		return true, PermError, errInvalidDomain
+		return true, PermError, ErrInvalidDomain
 	}
 
 	if r.ipNames == nil {
 		r.ipNames = []string{}
 		r.count++
-		ns, err := lookupAddr(r.ip.String())
+		ns, err := r.resolver.LookupAddr(r.ctx, r.ip.String())
 		if err != nil {
 			// https://tools.ietf.org/html/rfc7208#section-5
 			if isTemporary(err) {
@@ -387,15 +511,15 @@ func (r *resolution) ptrField(res Result, field, domain string) (bool, Result, e
 			// have some A/AAAA.
 			// https://tools.ietf.org/html/rfc7208#section-5.5
 			if r.count > 10 {
-				return false, "", errLookupLimitReached
+				return false, "", ErrLookupLimitReached
 			}
 			r.count++
-			addrs, err := lookupIP(n)
+			addrs, err := r.resolver.LookupIPAddr(r.ctx, n)
 			if err != nil {
 				// RFC explicitly says to skip domains which error here.
 				continue
 			}
-			trace("ptr forward resolution %q -> %q", n, addrs)
+			r.trace("ptr forward resolution %q -> %q", n, addrs)
 			if len(addrs) > 0 {
 				// Append the lower-case variants so we do a case-insensitive
 				// lookup below.
@@ -404,11 +528,11 @@ func (r *resolution) ptrField(res Result, field, domain string) (bool, Result, e
 		}
 	}
 
-	trace("ptr evaluating %q in %q", ptrDomain, r.ipNames)
+	r.trace("ptr evaluating %q in %q", ptrDomain, r.ipNames)
 	ptrDomain = strings.ToLower(ptrDomain)
 	for _, n := range r.ipNames {
 		if strings.HasSuffix(n, ptrDomain+".") {
-			return true, res, errMatchedPTR
+			return true, res, ErrMatchedPTR
 		}
 	}
 
@@ -422,15 +546,15 @@ func (r *resolution) existsField(res Result, field, domain string) (bool, Result
 	eDomain := field[7:]
 	eDomain, err := r.expandMacros(eDomain, domain)
 	if err != nil {
-		return true, PermError, errInvalidMacro
+		return true, PermError, ErrInvalidMacro
 	}
 
 	if eDomain == "" {
-		return true, PermError, errInvalidDomain
+		return true, PermError, ErrInvalidDomain
 	}
 
 	r.count++
-	ips, err := lookupIP(eDomain)
+	ips, err := r.resolver.LookupIPAddr(r.ctx, eDomain)
 	if err != nil {
 		// https://tools.ietf.org/html/rfc7208#section-5
 		if isTemporary(err) {
@@ -441,8 +565,8 @@ func (r *resolution) existsField(res Result, field, domain string) (bool, Result
 
 	// Exists only counts if there are IPv4 matches.
 	for _, ip := range ips {
-		if ip.To4() != nil {
-			return true, res, errMatchedExists
+		if ip.IP.To4() != nil {
+			return true, res, ErrMatchedExists
 		}
 	}
 	return false, "", nil
@@ -454,7 +578,7 @@ func (r *resolution) includeField(res Result, field, domain string) (bool, Resul
 	incdomain := field[len("include:"):]
 	incdomain, err := r.expandMacros(incdomain, domain)
 	if err != nil {
-		return true, PermError, errInvalidMacro
+		return true, PermError, ErrInvalidMacro
 	}
 	ir, err := r.Check(incdomain)
 	switch ir {
@@ -470,65 +594,62 @@ func (r *resolution) includeField(res Result, field, domain string) (bool, Resul
 		return true, PermError, err
 	}
 
-	return false, "", fmt.Errorf("This should never be reached")
+	return false, "", fmt.Errorf("this should never be reached")
 }
 
 type dualMasks struct {
-	v4 int
-	v6 int
+	v4 net.IPMask
+	v6 net.IPMask
 }
 
-func ipMatch(ip, tomatch net.IP, masks dualMasks) (bool, error) {
-	mask := -1
-	if tomatch.To4() != nil && masks.v4 >= 0 {
+func ipMatch(ip, tomatch net.IP, masks dualMasks) bool {
+	mask := net.IPMask(nil)
+	if tomatch.To4() != nil && masks.v4 != nil {
 		mask = masks.v4
-	} else if tomatch.To4() == nil && masks.v6 >= 0 {
+	} else if tomatch.To4() == nil && masks.v6 != nil {
 		mask = masks.v6
 	}
 
-	if mask >= 0 {
-		_, ipnet, err := net.ParseCIDR(
-			fmt.Sprintf("%s/%d", tomatch.String(), mask))
-		if err != nil {
-			return false, errInvalidMask
-		}
-		return ipnet.Contains(ip), nil
+	if mask != nil {
+		ipnet := net.IPNet{IP: tomatch, Mask: mask}
+		return ipnet.Contains(ip)
 	}
 
-	return ip.Equal(tomatch), nil
+	return ip.Equal(tomatch)
 }
 
 var aRegexp = regexp.MustCompile(`^[aA](:([^/]+))?(/(\w+))?(//(\w+))?$`)
 var mxRegexp = regexp.MustCompile(`^[mM][xX](:([^/]+))?(/(\w+))?(//(\w+))?$`)
 
 func domainAndMask(re *regexp.Regexp, field, domain string) (string, dualMasks, error) {
-	masks := dualMasks{-1, -1}
+	masks := dualMasks{}
 	groups := re.FindStringSubmatch(field)
 	if groups != nil {
 		if groups[2] != "" {
 			domain = groups[2]
 		}
 		if groups[4] != "" {
-			mask4, err := strconv.Atoi(groups[4])
-			if err != nil || mask4 < 0 || mask4 > 32 {
-				return "", masks, errInvalidMask
+			i, err := strconv.Atoi(groups[4])
+			mask4 := net.CIDRMask(i, 32)
+			if err != nil || mask4 == nil {
+				return "", masks, ErrInvalidMask
 			}
 			masks.v4 = mask4
 		}
 		if groups[6] != "" {
-			mask6, err := strconv.Atoi(groups[6])
-			if err != nil || mask6 < 0 || mask6 > 128 {
-				return "", masks, errInvalidMask
+			i, err := strconv.Atoi(groups[6])
+			mask6 := net.CIDRMask(i, 128)
+			if err != nil || mask6 == nil {
+				return "", masks, ErrInvalidMask
 			}
 			masks.v6 = mask6
 		}
 	}
-	trace("masks on %q: %q %q %v", field, groups, domain, masks)
 
 	// Test to catch malformed entries: if there's a /, there must be at least
 	// one mask.
-	if strings.Contains(field, "/") && masks.v4 == -1 && masks.v6 == -1 {
-		return "", masks, errInvalidMask
+	if strings.Contains(field, "/") && masks.v4 == nil && masks.v6 == nil {
+		return "", masks, ErrInvalidMask
 	}
 
 	return domain, masks, nil
@@ -538,16 +659,17 @@ func domainAndMask(re *regexp.Regexp, field, domain string) (string, dualMasks,
 func (r *resolution) aField(res Result, field, domain string) (bool, Result, error) {
 	// https://tools.ietf.org/html/rfc7208#section-5.3
 	aDomain, masks, err := domainAndMask(aRegexp, field, domain)
+	r.trace("masks on %q, %q: %q %v", field, domain, aDomain, masks)
 	if err != nil {
 		return true, PermError, err
 	}
 	aDomain, err = r.expandMacros(aDomain, domain)
 	if err != nil {
-		return true, PermError, errInvalidMacro
+		return true, PermError, ErrInvalidMacro
 	}
 
 	r.count++
-	ips, err := lookupIP(aDomain)
+	ips, err := r.resolver.LookupIPAddr(r.ctx, aDomain)
 	if err != nil {
 		// https://tools.ietf.org/html/rfc7208#section-5
 		if isTemporary(err) {
@@ -556,12 +678,9 @@ func (r *resolution) aField(res Result, field, domain string) (bool, Result, err
 		return false, "", err
 	}
 	for _, ip := range ips {
-		ok, err := ipMatch(r.ip, ip, masks)
-		if ok {
-			trace("mx matched %v, %v, %v", r.ip, ip, masks)
-			return true, res, errMatchedA
-		} else if err != nil {
-			return true, PermError, err
+		if ipMatch(r.ip, ip.IP, masks) {
+			r.trace("a matched %v, %v, %v", r.ip, ip.IP, masks)
+			return true, res, ErrMatchedA
 		}
 	}
 
@@ -572,16 +691,17 @@ func (r *resolution) aField(res Result, field, domain string) (bool, Result, err
 func (r *resolution) mxField(res Result, field, domain string) (bool, Result, error) {
 	// https://tools.ietf.org/html/rfc7208#section-5.4
 	mxDomain, masks, err := domainAndMask(mxRegexp, field, domain)
+	r.trace("masks on %q, %q: %q %v", field, domain, mxDomain, masks)
 	if err != nil {
 		return true, PermError, err
 	}
 	mxDomain, err = r.expandMacros(mxDomain, domain)
 	if err != nil {
-		return true, PermError, errInvalidMacro
+		return true, PermError, ErrInvalidMacro
 	}
 
 	r.count++
-	mxs, err := lookupMX(mxDomain)
+	mxs, err := r.resolver.LookupMX(r.ctx, mxDomain)
 	if err != nil {
 		// https://tools.ietf.org/html/rfc7208#section-5
 		if isTemporary(err) {
@@ -593,13 +713,13 @@ func (r *resolution) mxField(res Result, field, domain string) (bool, Result, er
 	// There's an explicit maximum of 10 MX records per match.
 	// https://tools.ietf.org/html/rfc7208#section-4.6.4
 	if len(mxs) > 10 {
-		return true, PermError, errTooManyMXRecords
+		return true, PermError, ErrTooManyMXRecords
 	}
 
 	mxips := []net.IP{}
 	for _, mx := range mxs {
 		r.count++
-		ips, err := lookupIP(mx.Host)
+		ips, err := r.resolver.LookupIPAddr(r.ctx, mx.Host)
 		if err != nil {
 			// https://tools.ietf.org/html/rfc7208#section-5
 			if isTemporary(err) {
@@ -607,15 +727,14 @@ func (r *resolution) mxField(res Result, field, domain string) (bool, Result, er
 			}
 			return false, "", err
 		}
-		mxips = append(mxips, ips...)
+		for _, ipaddr := range ips {
+			mxips = append(mxips, ipaddr.IP)
+		}
 	}
 	for _, ip := range mxips {
-		ok, err := ipMatch(r.ip, ip, masks)
-		if ok {
-			trace("mx matched %v, %v, %v", r.ip, ip, masks)
-			return true, res, errMatchedMX
-		} else if err != nil {
-			return true, PermError, err
+		if ipMatch(r.ip, ip, masks) {
+			r.trace("mx matched %v, %v, %v", r.ip, ip, masks)
+			return true, res, ErrMatchedMX
 		}
 	}
 
@@ -627,11 +746,11 @@ func (r *resolution) redirectField(field, domain string) (Result, error) {
 	rDomain := field[len("redirect="):]
 	rDomain, err := r.expandMacros(rDomain, domain)
 	if err != nil {
-		return PermError, errInvalidMacro
+		return PermError, ErrInvalidMacro
 	}
 
 	if rDomain == "" {
-		return PermError, errInvalidDomain
+		return PermError, ErrInvalidDomain
 	}
 
 	// https://tools.ietf.org/html/rfc7208#section-6.1
@@ -656,8 +775,8 @@ func (r *resolution) expandMacros(s, domain string) (string, error) {
 	// from happening in case where it matters (a, mx), but for the ones which
 	// doesn't, prevent them from sneaking through.
 	if strings.Contains(s, "/") {
-		trace("macro contains /")
-		return "", errInvalidDomain
+		r.trace("macro contains /")
+		return "", ErrInvalidDomain
 	}
 
 	// Bypass the complex logic if there are no macros present.
@@ -693,7 +812,7 @@ func (r *resolution) expandMacros(s, domain string) (string, error) {
 				inMacroDefinition = true
 				continue
 			}
-			return "", errInvalidMacro
+			return "", ErrInvalidMacro
 		}
 		if inMacroDefinition {
 			if c != '}' {
@@ -705,10 +824,10 @@ func (r *resolution) expandMacros(s, domain string) (string, error) {
 			// Extract letter, digit transformer, reverse transformer, and
 			// delimiters.
 			groups := macroRegexp.FindStringSubmatch(macroS)
-			trace("macro %q: %q", macroS, groups)
+			r.trace("macro %q: %q", macroS, groups)
 			macroS = ""
 			if groups == nil {
-				return "", errInvalidMacro
+				return "", ErrInvalidMacro
 			}
 			letter := groups[1]
 
@@ -718,7 +837,7 @@ func (r *resolution) expandMacros(s, domain string) (string, error) {
 				// valid.
 				digits, err = strconv.Atoi(groups[2])
 				if err != nil || digits <= 0 {
-					return "", errInvalidMacro
+					return "", ErrInvalidMacro
 				}
 			}
 			reverse := groups[3] == "r" || groups[3] == "R"
@@ -743,7 +862,7 @@ func (r *resolution) expandMacros(s, domain string) (string, error) {
 			case "d":
 				str = domain
 			case "i":
-				str = r.ip.String()
+				str = ipToMacroStr(r.ip)
 			case "p":
 				// This shouldn't be used, we don't want to support it, it's
 				// risky. "unknown" is a safe value.
@@ -756,11 +875,11 @@ func (r *resolution) expandMacros(s, domain string) (string, error) {
 					str = "ip6"
 				}
 			case "h":
-				str = domain
+				str = r.helo
 			default:
 				// c, r, t are allowed in exp only, and we don't expand macros
 				// in exp so they are just as invalid as the rest.
-				return "", errInvalidMacro
+				return "", ErrInvalidMacro
 			}
 
 			// Split str using the given separators.
@@ -802,7 +921,7 @@ func (r *resolution) expandMacros(s, domain string) (string, error) {
 		n += string(c)
 	}
 
-	trace("macro expanded %q to %q", s, n)
+	r.trace("macro expanded %q to %q", s, n)
 	return n, nil
 }
 
@@ -811,3 +930,19 @@ func reverseStrings(a []string) {
 		a[left], a[right] = a[right], a[left]
 	}
 }
+
+func ipToMacroStr(ip net.IP) string {
+	if ip.To4() != nil {
+		return ip.String()
+	}
+
+	// For IPv6 addresses, the "i" macro expands to a dot-format address.
+	// https://datatracker.ietf.org/doc/html/rfc7208#section-7.3
+	sb := strings.Builder{}
+	sb.Grow(64)
+	for _, b := range ip.To16() {
+		fmt.Fprintf(&sb, "%x.%x.", b>>4, b&0xf)
+	}
+	// Return the string without the trailing ".".
+	return sb.String()[:sb.Len()-1]
+}
